#include <string.h>
#include <dos.h>

#define rgb15(r,g,b) (unsigned(b) | (unsigned(g) << 5u) | (unsigned(r) << 10u))

#define tovga(c)  (((unsigned char)(c) << 1u) | ((unsigned char)(c) >> 4u)) /* r*63/31 */

/* VGA color number to VGA DAC attribute number mapping. */
static const unsigned char indexes[16] =
    { 0, 1, 2, 3, 4, 5,20, 7,
     56,57,58,59,60,61,62,63};

/* Mapping of DAC attribute numbers to original VGA colors */
static const unsigned int origmap256[16] =
{
    rgb15( 0, 0, 0), rgb15( 0, 0,21), rgb15( 0,21, 0), rgb15( 0,21,21),
    rgb15(21, 0, 0), rgb15(21, 0,21), rgb15(21, 5, 0), rgb15(21,21,21),
    rgb15( 7, 7, 7), rgb15( 5, 5,31), rgb15( 5,31, 5), rgb15( 5,31,31),
    rgb15(31, 5, 5), rgb15(31, 5,31), rgb15(31,31, 5), rgb15(31,31,31)
};

/*
   VGA colors:

   Extended colors:
    6x6x6 cube with following values:
               n ? 55 + n*40 : 0
       OR VGA:
               n ? 13 + n*10 : 0
    24 grays with following values:
               8 + n*10
       OR:
               (n + (n+1) << 1) >> 1
 */


/* Currently mapped color in each attribute, in rgb() format.
 * If bit 0x8000 is set, the color is unchanged from original.
 */
static unsigned int map256[16];

static unsigned char recent[16];
int Find256Color(unsigned color)
{
    signed char r,g,b;
    if(color < 16)
    {
        static const unsigned char swap8[8] = {0,4,2,6,1,5,3,7};
        color = (color&8) | swap8[color&7];
        if(!(map256[color] & 0x8000u))
        {
            return color;
        }
        // Colors 0..15 map into real VGA colors.
        color = origmap256[color];
        b = color & 0x1F;
        g = (color >> 5) & 0x1F;
        r = (color >> 10) & 0x1F;
    }
    else if(color < 232)
    {
        static const unsigned char ramp[6] = { 0,12,16,21,26,31 };
        color -= 16;
        b = ramp[ color % 6u ];
        g = ramp[ (color / 6u) % 6u ];
        r = ramp[ (color / 36u) % 6u ];
        color = rgb15(r,g,b);
    }
    else
    {
        static const unsigned char ramp[24] = { 1,2,3,5,6,7,8,9,11,12,13,14,16,17,18,19,20,22,23,24,25,27,28,29 };
        r = g = b = ramp[ color - 232 ];
        color = rgb15(r,g,b);
    }

    unsigned bestindex = 15, bestscore = 32767;

    for(unsigned c=0; c<16; ++c) // Start from 1. We don't change black.
    {
        // Which color is _currently_ in this slot?
        unsigned curcolor = map256[c] & 0x7FFF;
        if(curcolor == color) return c;
        if(c==0) continue;

        // How much does it differ from the desired color?
        unsigned score = 0;
        /**/
        { int diff = int(b) - int(curcolor & 0x1F);         score += diff*diff; }
        { int diff = int(g) - int((curcolor >> 5) & 0x1F);  score += diff*diff; }
        { int diff = int(r) - int((curcolor >> 10) & 0x1F); score += diff*diff; }
        /*
        { signed char bdiff = (b) - (curcolor & 0x1F);         score += bdiff*bdiff; }
        { signed char gdiff = (g) - ((curcolor >> 5) & 0x1F);  score += gdiff*gdiff; }
        { signed char rdiff = (r) - ((curcolor >> 10) & 0x1F); score += rdiff*rdiff; }
        */
        // max score: 31*31 * 3 = 2883

        //if(!(map256[c] & 0x8000)) { int tmp = 16-c; if(score > tmp) score = tmp; }

        // If we _just_ changed this color, add penalty
        if(recent[c]) score += 1000 + recent[c]*100;

        if(score < bestscore) { bestscore = score; bestindex = c; }
    }

    // Install it
    outportb(0x3C8, indexes[bestindex]);
    outportb(0x3C9, tovga(r));
    outportb(0x3C9, tovga(g));
    outportb(0x3C9, tovga(b));

    // Bookkeeping: This RGB color is now residing in this index.
    map256[bestindex] = color | 0x8000u;

finalize:
    for(unsigned n=0; n<16; ++n)
        if(n == bestindex) recent[n] = 16;
        else if(recent[n]) --recent[n];
    return bestindex;
}
void Reset256Color()
{
    for(unsigned c=0; c<16; ++c)
    {
        // Load RGB color
        unsigned m = origmap256[c]; map256[c] = m;
        // Decompose it
        unsigned char b = m & 0x1F;
        unsigned char g = (m >> 5) & 0x1F;
        unsigned char r = (m >> 10) & 0x1F;
        // Install
        outportb(0x3C8, indexes[c]);
        outportb(0x3C9, tovga(r));
        outportb(0x3C9, tovga(g));
        outportb(0x3C9, tovga(b));
    }
}

unsigned Verify256Color(int color, int hint_we_expect)
{
    // For given colorindex (0-15), returns the "256-color" index currently in that palette slot.
    recent[color] = 15;
    unsigned what = map256[color];
    if(!what & 0x8000u)
    {
        // Unchanged VGA color
        return color;
    }
    // Decompose it
    unsigned char b = what & 0x1F;
    unsigned char g = (what >> 5) & 0x1F;
    unsigned char r = (what >> 10) & 0x1F;
    if(r == g && r == b)
    {
        // Gray (could still be colored though)
        if(r < 22)
            if(r < 10)
                if(r <= 4)
                    if(!r || r==4) goto colored;
                    else return 232+r-1;      // 1..3 = 0..2
                else return 232+r-2;          // 4..9 = 3..7
            else if(r == 10 || r == 15) goto colored;
            else if(r < 15) return 232+r-3;  // 11..14 = 4..11
            else if(r == 21) goto colored;
            else return 232+r-4;             // 16..20 = 12..16
        else if(r < 26) return 232+r-5;      // 22..25 = 17..20
        else if(r == 26 || r >= 30) goto colored;
        else return 232+r-6;                 // 27..29 = 21..23
    }
    else
    {
    colored:;
        // Colored
        if(r) r = (r-1)/6;
        if(g) g = (r-1)/6;
        if(b) b = (r-1)/6;
        return 16 + b + g*6 + r*36;
    }
}

#undef rgb

int Find256Color_DontChange(unsigned color)
{
    signed char r,g,b;
    if(color < 16)
    {
        static const unsigned char swap8[8] = {0,4,2,6,1,5,3,7};
        color = (color&8) | swap8[color&7];
        if(!(map256[color] & 0x8000u))
        {
            return color;
        }
        // Colors 0..15 map into real VGA colors.
        color = origmap256[color];
        b = color & 0x1F;
        g = (color >> 5) & 0x1F;
        r = (color >> 10) & 0x1F;
    }
    else if(color < 232)
    {
        static const unsigned char ramp[6] = { 0,12,16,21,26,31 };
        color -= 16;
        b = ramp[ color % 6u ];
        g = ramp[ (color / 6u) % 6u ];
        r = ramp[ (color / 36u) % 6u ];
        color = rgb15(r,g,b);
    }
    else
    {
        static const unsigned char ramp[24] = { 1,2,3,5,6,7,8,9,11,12,13,14,16,17,18,19,20,22,23,24,25,27,28,29 };
        r = g = b = ramp[ color - 232 ];
        color = rgb15(r,g,b);
    }

    unsigned bestindex = 15, bestscore = 32767;

    for(unsigned c=0; c<16; ++c) // Start from 1. We don't change black.
    {
        // Which color is _currently_ in this slot?
        unsigned curcolor = map256[c] & 0x7FFF;
        if(curcolor == color) return c;
        if(c==0) continue;

        // How much does it differ from the desired color?
        unsigned score = 0;
        { int diff = int(b) - int(curcolor & 0x1F);         score += diff*diff; }
        { int diff = int(g) - int((curcolor >> 5) & 0x1F);  score += diff*diff; }
        { int diff = int(r) - int((curcolor >> 10) & 0x1F); score += diff*diff; }
        if(score < bestscore) { bestscore = score; bestindex = c; }
    }
    return bestindex;
}
